#define ServerAddress "127.0.0.1" // For now, it doesn't have any configuration, so
                                  // change the server address here. It can be the
                                  // computer name.

#include <Alert.h>
#include <Application.h>
#include <Bitmap.h>
#include <BitmapStream.h>
#include <Locker.h>
#include <PFZNetEndpointClient.h>
#include <Screen.h>
#include <TranslatorRoster.h>
#include <View.h>
#include <Window.h>

#include <stdio.h>

#define ShowMessage(msg) ((new BAlert(msg,msg,"OK"))->Go())

// The epDevice is used to administrate a remote machine. But the code is unstable
// for now. (And requires that you install an add-on to the input_server, but I don't
// even send this add-on for now).

class pfzWindow;
class pfzView;

class pfzClient;

class pfzApplication:public BApplication{
 private:
  typedef BApplication inherited;
 public:
  BTranslatorRoster *roster;
  BBitmap *image;
  BLocker imageLocker;
  pfzWindow *win;
  pfzClient *ep;
//  TPFZNetEndpoint *epDevice;
  pfzApplication();
  ~pfzApplication();
  bool QuitRequested(void);
};

class pfzWindow:public BWindow{
 typedef BWindow inherited;
 public:
  pfzView *view;
  pfzWindow(BRect Bounds);
  ~pfzWindow();
  virtual void MessageReceived(BMessage *msg);
  bool QuitRequested(void);
};

class pfzView:public BView{
 typedef BView inherited;
 public:
//  bool isDown;
  pfzView(BRect bounds);
  void AttachedToWindow(void);
  void Draw(BRect updateRect);
/*  virtual void MouseDown(BPoint point);
  virtual void MouseUp(BPoint point);
  virtual void MouseMoved(BPoint point,uint32 transit,const BMessage *message);
  virtual void KeyDown(const char *bytes,int32 numBytes);
  virtual void KeyUp(const char *bytes,int32 numBytes);
  void sendInputMessage();
  void sendKeyInputMessage();*/
};

class pfzClient:public TPFZNetEndpointClient{
 public:
  int32 OnRun();
  void OnStopRunning();
};

sem_id globalSem=-1;

int main(void){
pfzApplication app;
 app.Run();
 return 0;
}

pfzApplication::pfzApplication():BApplication("application/x-vnd.PFZ-DesktopReceiver"){
 try{
  roster=BTranslatorRoster::Default();
  // This is the roster used to convert the image received from the network to a
  // bitmap. The image is in JPEG (considering you are using my server), but it can
  // read any image that Translation Kit can read.
  ep=new pfzClient;
//  epDevice=new TPFZNetEndpoint;
  if (ep->InitCheck()==B_OK){
  BScreen scr;
  BRect bounds=scr.Frame();
  int32 x,y;
   x=(int32)(bounds.right/2);
   y=(int32)(bounds.bottom/2);
   new pfzWindow(BRect(x-200,y-150,x+199,y+149));
   return;
  }
 } catch(...){
 }
 ShowMessage("Error starting.");
 PostMessage(B_QUIT_REQUESTED);
}

pfzApplication::~pfzApplication(){
// delete epDevice;
 delete ep;
}

bool pfzApplication::QuitRequested(){
bool result;
 result=inherited::QuitRequested();
 if (result){
  if (win){
   Unlock();
   win->PostMessage(B_QUIT_REQUESTED);
   acquire_sem(globalSem); // Wait until this semaphore is released to really quit the
                           // the application.
   Lock();
  }
 }
 return result;
}

#define app ((pfzApplication *)be_app)
pfzWindow::pfzWindow(BRect bounds):inherited(bounds,"Desktop Receiver",B_TITLED_WINDOW,B_ASYNCHRONOUS_CONTROLS){
 app->win=this;
 AddChild(view=new pfzView(Bounds()));
 if (!app->ep->Run()){
  ShowMessage("Can't start receiverThread.");
  PostMessage(B_QUIT_REQUESTED);
 }
 BMessage *shortcutMsg=new BMessage('PFZf');
 AddShortcut('f',B_SHIFT_KEY|B_CONTROL_KEY,shortcutMsg);
// Press Ctrl+Alt+Shift+F for full screen. This key combination is this complex
// because when I start administrating remote machines, Alt+F and Ctrl+F will be sent
// to that machine.
 globalSem=create_sem(0,"globalSem");
 Show();
}

pfzWindow::~pfzWindow(){
}

bool pfzWindow::QuitRequested(){
 Unlock(); // Unlock so the Client thread doesn't wait to lock the window while we wait
           // for it to close (or we will wait forever). More simple, avoid dead locks.
 be_app->PostMessage(B_QUIT_REQUESTED);
 app->ep->Close();
 app->ep->WaitForConnectionClose();
 release_sem(globalSem); // Now the application can REALLY quit.
 Lock();
 return inherited::QuitRequested();
}

void pfzWindow::MessageReceived(BMessage *msg){
 switch(msg->what){
  case 'PFZf': {
  BScreen scr;
  BRect bounds=scr.Frame();
  BLooper *locker=Looper();
   if (locker->Lock()){
    if (bounds==Frame()){
     MoveTo((bounds.right/2)-200,(bounds.bottom/2)-150);
     ResizeTo(400,300);
    } else {
     MoveTo(0,0);
     ResizeTo(bounds.right,bounds.bottom);
    }
    locker->Unlock();
   }
   break;
  }
  default:
   inherited::MessageReceived(msg);
 }
}

pfzView::pfzView(BRect bounds):BView(bounds,"DesktopReceiverView",B_FOLLOW_ALL_SIDES,B_WILL_DRAW|B_FULL_UPDATE_ON_RESIZE){
}

void pfzView::AttachedToWindow(){
 SetHighColor(ui_color(B_PANEL_BACKGROUND_COLOR));
 SetViewColor(B_TRANSPARENT_COLOR);
 SetEventMask(B_NO_POINTER_HISTORY); // You can ignore this, as we are not sending the
                                     // mouse events to the network.
 MakeFocus();
}

void pfzView::Draw(BRect updateRect){
 app->imageLocker.Lock();
 if (app->image) {
  DrawBitmap(app->image,Bounds());
 } else {
  FillRect(updateRect);
 }
 app->imageLocker.Unlock();
}

/*void pfzView::MouseDown(BPoint point){
 isDown=true;
 sendInputMessage();
 inherited::MouseDown(point);
}
void pfzView::MouseUp(BPoint point){
 isDown=false;
 sendInputMessage();
 inherited::MouseUp(point);
}
void pfzView::MouseMoved(BPoint point,uint32 transit,const BMessage *message){
 if (isDown){
  sendInputMessage();
 }
 inherited::MouseMoved(point,transit,message);
}

void pfzView::KeyDown(const char *bytes,int32 numBytes){
 sendKeyInputMessage();
 inherited::KeyDown(bytes,numBytes);
}

void pfzView::KeyUp(const char *bytes,int32 numBytes){
 sendKeyInputMessage();
 inherited::KeyUp(bytes,numBytes);
}

void pfzView::sendInputMessage(){
TPFZNetEndpoint *ep=app->epDevice;
BMessage *msg=Window()->CurrentMessage();
 if (ep->SendInt32(msg->what)){
  BPoint where;
  if (msg->FindPoint("where",&where)==B_OK){
  int32 x,y;
  BRect bounds;
   bounds=Bounds();
   x=(int32)(where.x*1000000000/bounds.right);
   y=(int32)(where.y*1000000000/bounds.bottom);
   if (ep->SendInt32(x)){
    if (ep->SendInt32(y)){
    int32 modifiers;
     if (msg->FindInt32("modifiers",&modifiers)==B_OK){
      if (ep->SendInt32(modifiers)){
      int32 buttons;
       if (msg->what!=B_MOUSE_DOWN) return;
       if (msg->FindInt32("buttons",&buttons)==B_OK){
        if (ep->SendInt32(buttons)){
         return;
        }
       }
      }
     }
    }
   }
  }
 }
 ep->Close();
}

void pfzView::sendKeyInputMessage(){
TPFZNetEndpoint *ep=app->epDevice;
BMessage *msg=Window()->CurrentMessage();
 if (ep->SendInt32(msg->what)){
 int8 byte;
  if (msg->FindInt8("byte",&byte)==B_OK){
   if (ep->SendInt8(byte)){
   int32 modifiers;
    if (msg->FindInt32("modifiers",&modifiers)==B_OK){
     if (ep->SendInt32(modifiers)){
      return;
     }
    }
   }
  }
 }
 ep->Close();
}*/

int32 pfzClient::OnRun(){
char Quality,Qualities;
BMallocIO *Mem=NULL;
BRect rect;
int32 W,H,i;
pfzWindow *win=app->win;
pfzView *view=win->view;
BLooper *viewLooper=view->Looper();
 try{
  if (Connect(ServerAddress,3210)==B_OK){
/*   if (epDevice->Connect(ServerAddress,3211)==B_OK){
    ShowMessage("You started as the administrator of that machine.");
   }*/
   try{
    Mem=new BMallocIO;
    if (GetData(&Quality,1)){ // All these quality attributes are not used for now
     if (GetData(&Qualities,1)){ // in the BeOS implementation (I have a program similar
      if (GetData(&Qualities,1)){ // to this in Delphi, and it uses the Qualities.
       for (;;){
        app->win->Lock();
        rect=app->win->Bounds();
        app->win->Unlock();
        W=(int32)rect.right+1;
        H=(int32)rect.bottom+1;
        if (!SendData(&Quality,1)) break;
        if (!SendInt32(W)) break;
        if (!SendInt32(H)) break;
        if (!GetInt32(&i)) break;
        Mem->SetSize(i);
        if (!GetData((void *)Mem->Buffer(),i)) break;
        BBitmapStream *stream;
        stream=new BBitmapStream;
        if (app->roster->Translate(Mem,NULL,NULL,stream,B_TRANSLATOR_BITMAP)<B_OK){
         delete stream;
         break;
        }
        app->imageLocker.Lock();
        delete app->image;
        stream->DetachBitmap(&app->image);
        app->imageLocker.Unlock();
        delete stream;
        viewLooper->Lock();
        view->Invalidate();
        viewLooper->Unlock();
        Mem->Seek(0,SEEK_SET);
        Mem->SetSize(0);
       }
      }
     }
    }
   } catch(...){
   }
   delete Mem;
  } else {
   ShowMessage("Can't connect to the server.");
  }
 } catch(...){
 }
 return 0;
}

void pfzClient::OnStopRunning(){
 be_app->PostMessage(B_QUIT_REQUESTED);
 // When the connection close, close the application too.
}

#undef app