#include <Alert.h>
#include <Application.h>
#include <Bitmap.h>
#include <BitmapStream.h>
#include <DataIO.h>
#include <Locker.h>
#include <PFZNetEndpointServer.h>
#include <Screen.h>
#include <TranslatorRoster.h>
#include <View.h>
#include <stdio.h>

#define ShowMessage(msg) ((new BAlert(msg,msg,"OK"))->Go())

class pfzServer:public TPFZNetEndpointServer{
 public:
  void OnServerStopRunning();
  int32 OnClientRun(void *data);
};

class pfzApplication:public BApplication{
 private:
  typedef BApplication inherited;
 public:
  BTranslatorRoster *roster;
  pfzServer *ServerEP;
  pfzApplication();
  ~pfzApplication();
  bool QuitRequested();
};

char Quality=30;
char MinQuality=1;
char MaxQuality=100;

int main(void){
pfzApplication app;
 app.Run();
 return 0;
}

#define app ((pfzApplication *)be_app)

pfzApplication::pfzApplication():BApplication("application/x-vnd.PFZ-DesktopSender"){
int16 port=3210;
 try{
  roster=BTranslatorRoster::Default();
  ServerEP=new pfzServer;
  if (ServerEP->InitCheck()==B_OK){
   if (ServerEP->Bind(port)==B_OK){
    if (ServerEP->Listen()==B_OK){
     if (ServerEP->Run()){
      return;
     }
    }
   }
  }
 } catch(...){
 }
 ShowMessage("Error initializing the server.");
 PostMessage(B_QUIT_REQUESTED);
}

pfzApplication::~pfzApplication(){
 delete ServerEP;
}

bool CloseClientsDoForEach(void *ThreadData){
#define ep ((TPFZNetEndpoint *)ThreadData)
 ep->Close();
 return false;
#undef data
}

bool pfzApplication::QuitRequested(){
 if (ServerEP){
  Unlock(); // avoid dead locks.
  ServerEP->Run(false);
  ServerEP->WaitForConnectionsClose();
  Lock();
 }
 return inherited::QuitRequested();
}

void pfzServer::OnServerStopRunning(){
 be_app->PostMessage(B_QUIT_REQUESTED); // The server has stopped, so we quit.
}

int32 pfzServer::OnClientRun(void *ThreadData){
#define ep ((TPFZNetEndpoint *)ThreadData)
BScreen Screen;
BBitmap *bmp,*jpg;
char B;
int32 W,H,bmpW,bmpH;
BRect bmpBounds,bounds(0,0,0,0);
BView *view=NULL;
BMallocIO *Mem=NULL;
BBitmapStream *stream;
int32 Size;
  try{
   Mem=new BMallocIO;
   Mem->SetBlockSize(8*1024);
   view=new BView(BRect(0,0,0,0),"viewForCreatingJpgs",B_FOLLOW_ALL_SIDES,B_WILL_DRAW);
   if (ep->SendData(&Quality,1)){ // Send our default quality configuration. (Unused for now)
    if (ep->SendData(&MinQuality,1)){
     if (ep->SendData(&MaxQuality,1)){
      for (;;){
       if (!ep->GetData(&B,sizeof(B))) break;
       if (!ep->GetInt32(&W)) break;
       if (!ep->GetInt32(&H)) break;
       if (Screen.GetBitmap(&bmp,true)!=B_OK) break;
       bmpBounds=bmp->Bounds();
       bmpW=((int32)(bmpBounds.right-bmpBounds.left))+1;
       bmpH=((int32)(bmpBounds.bottom-bmpBounds.top))+1;
       jpg=bmp;
       if (W>bmpW) {
        W=bmpW;
       } else if (W<bmpW){
        jpg=NULL;
       }
       if (H>bmpH){
        H=bmpH;
       } else if (H<bmpH){
        jpg=NULL;
       }
       if (!jpg){
        bounds.right=W-1;
        bounds.bottom=H-1;
        jpg=new BBitmap(bounds,bmp->ColorSpace(),true);
        view->ResizeTo(W,H);
        jpg->AddChild(view);
        jpg->Lock();
        view->DrawBitmap(bmp,bounds);
        jpg->Unlock();
        jpg->RemoveChild(view);
        delete bmp;
        bmp=jpg;
       }
       stream=new BBitmapStream(bmp);
       app->roster->Translate(stream,NULL,NULL,Mem,B_JPEG_FORMAT);
       delete stream;
       Size=Mem->BufferLength();
       if (!ep->SendInt32(Size)) break;
       if (!ep->SendData(Mem->Buffer(),Size)) break;
       Mem->Seek(0,SEEK_SET);
       Mem->SetSize(0);
      }
     }
    }
   }
   delete view;
   delete Mem;
  } catch(...){
  }
 return 0;
#undef ep
}
